;; -*- geiser-scheme-implementation: guile -*-

;; bullet.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td bullet)
  #:use-module (td hooks)
  #:use-module (td util)
  #:use-module (td euclidean)
  #:use-module (td drawable)
  #:use-module (td enemy)
  #:use-module (td tick)
  #:use-module (aa-tree)
  #:use-module (chickadee)
  #:use-module (oop goops)
  #:export (bullets
            launch-simple-bullet
            launch-fire-bullet
            launch-delay-bullet
            launch-damaging-delay-bullet
            process-bullet-procedures))

(define bullet-procedures #nil)
(define bullets #nil)

(add-hook! new-game-hook
           (lambda ()
             (set! bullet-procedures #nil)
             (set! bullets #nil)))

(define (launch-simple-bullet start-x start-y target-uid dmg)
  (let ((bullet-id (get-uid))
        (bullet-object (make <drawable>
                         #:x start-x #:y start-y
                         #:texture-proc (lambda (_) #:bullet))))
    ;; Register de bullet
    (set! bullets
      (aa-insert bullets
                 bullet-id
                 bullet-object))
    
    (let* ((target-enemy (aa-search enemies target-uid))
           (target-x (.x target-enemy))
           (target-y (.y target-enemy)))
      (let ((bullet-procedure
             (lambda ()
               (catch 'arrived

                 ;; Keep going...
                 (lambda ()
                   (set! bullets
                     (aa-modify bullets bullet-id
                                (lambda (old-bullet)
                                  (approach old-bullet
                                            (make <euclidean> #:x target-x #:y target-y)
                                            (* period 0.2))))))

                 ;; Got 'em!
                 (lambda (_)
                   (set! bullets (aa-delete bullets bullet-id))
                   (damage target-uid dmg)
                   (throw 'terminate))))))
        (set! bullet-procedures (cons bullet-procedure bullet-procedures))))

    bullet-id))

(define (launch-fire-bullet start-x start-y target-uid dmg)
  (let ((bullet-id (get-uid))
        (bullet-object (make <drawable>
                         #:x start-x #:y start-y
                         #:texture-proc (lambda (_) #:fire-bullet))))
    ;; Register de bullet
    (set! bullets
      (aa-insert bullets
                 bullet-id
                 bullet-object))
    
    (let* ((target-enemy (aa-search enemies target-uid))
           (target-x (.x target-enemy))
           (target-y (.y target-enemy)))
      (let ((bullet-procedure
             (lambda ()
               (catch 'arrived

                 ;; Keep going...
                 (lambda ()
                   (set! bullets
                     (aa-modify bullets bullet-id
                                (lambda (old-bullet)
                                  (approach old-bullet
                                            (make <euclidean> #:x target-x #:y target-y)
                                            (* period 0.2))))))

                 ;; Got 'em!
                 (lambda (_)
                   (set! bullets (aa-delete bullets bullet-id))
                   (damage target-uid dmg)
                   (throw 'terminate))))))
        (set! bullet-procedures (cons bullet-procedure bullet-procedures))))

    bullet-id))

(define (launch-delay-bullet start-x start-y delay-ms target-uid)
  (let ((bullet-id (get-uid))
        (bullet-object (make <drawable>
                         #:x start-x #:y start-y
                         #:texture-proc (lambda (_) #:white-bullet))))

    ;; Register de bullet
    (set! bullets
      (aa-insert bullets
                 bullet-id
                 bullet-object))
    
    (let* ((target-enemy (aa-search enemies target-uid))
           (target-x (.x target-enemy))
           (target-y (.y target-enemy)))
      (let ((bullet-procedure
             (lambda ()
               (catch 'arrived

                 ;; Keep going...
                 (lambda ()
                   (set! bullets
                     (aa-modify bullets bullet-id
                                (lambda (old-bullet)
                                  (approach old-bullet
                                            (make <euclidean> #:x target-x #:y target-y)
                                            (* period 0.2))))))

                 ;; Got 'em!
                 (lambda (_)
                   (set! bullets (aa-delete bullets bullet-id))
                   (slow target-uid delay-ms)
                   (throw 'terminate))))))
        (set! bullet-procedures (cons bullet-procedure bullet-procedures))))

    bullet-id))

(define (launch-damaging-delay-bullet start-x start-y delay-ms dmg target-uid)
  (let ((bullet-id (get-uid))
        (bullet-object (make <drawable>
                         #:x start-x #:y start-y
                         #:texture-proc (lambda (_) #:purple-bullet))))

    ;; Register de bullet
    (set! bullets
      (aa-insert bullets
                 bullet-id
                 bullet-object))
    
    (let* ((target-enemy (aa-search enemies target-uid))
           (target-x (.x target-enemy))
           (target-y (.y target-enemy)))
      (let ((bullet-procedure
             (lambda ()
               (catch 'arrived

                 ;; Keep going...
                 (lambda ()
                   (set! bullets
                     (aa-modify bullets bullet-id
                                (lambda (old-bullet)
                                  (approach old-bullet
                                            (make <euclidean> #:x target-x #:y target-y)
                                            (* period 0.2))))))

                 ;; Got 'em!
                 (lambda (_)
                   (set! bullets (aa-delete bullets bullet-id))
                   (slow target-uid delay-ms)
                   (damage target-uid dmg)
                   (throw 'terminate))))))
        (set! bullet-procedures (cons bullet-procedure bullet-procedures))))

    bullet-id))

(define (process-bullet-procedures)
  (let ((new-queue #nil))
    (for-each (lambda (p)
                (catch 'terminate
                  (lambda ()
                    (p)
                    (set! new-queue (cons p new-queue)))
                  (lambda (_) #nil)))
              bullet-procedures)
    (set! bullet-procedures new-queue)))

(add-hook! update-hook
           (lambda (dt)
             (unless (game-paused?)
               (process-bullet-procedures))))
