;; -*- geiser-scheme-implementation: guile -*-

;; euclidean.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(read-set! keywords 'prefix)

(define-module (td euclidean)
  :use-module (td fset)
  :use-module (aa-tree)
  :use-module (oop goops)
  :export (<euclidean>
           .x
           .y
           kind-of-close
           dist
           approach
           nearest))

(define-class <euclidean> ()
  (x :init-value 0 :accessor .x :init-keyword :x)
  (y :init-value 0 :accessor .y :init-keyword :y))

(define-method (kind-of-close (a <euclidean>) (b <euclidean>) range)
  (let ((dx (abs (- (.x b) (.x a))))
        (dy (abs (- (.y b) (.y a)))))
    (and (<= dx range)
         (<= dy range))))

(define-method (dist (a <euclidean>) (b <euclidean>))
  (let ((m (- (.x b) (.x a)))
        (n (- (.y b) (.y a))))
    (sqrt (+ (* m m) (* n n)))))

;;    /|
;; C / |
;;  /  | B
;; /___|
;;  A
(define-method (approach (m <euclidean>) (n <euclidean>) distance)
  (let ((c (dist m n)))
    (if (<= c distance)
        (throw 'arrived)
        (let* ((ratio (/ distance c))
               (a (- (.x n) (.x m)))
               (b (- (.y n) (.y m))))
          (clone m
                 ((.x) (+ (.x m) (* a ratio)))
                 ((.y) (+ (.y m) (* b ratio))))))))

(define-method (nearest (ref-e <euclidean>) id-tree range)
  (if (nil? id-tree)
      (throw 'no-candidate)
      (let ((closest-id #nil)
            (closest-dist #nil))
        (aa-for-each
         (lambda (p)
           (let ((id (car p))
                 (cand-e (cdr p)))
             (unless (not (kind-of-close ref-e cand-e range))
               (let ((d (dist ref-e cand-e)))
                 (when (or (nil? closest-id)
                           (< d closest-dist))
                   (set! closest-id id)
                   (set! closest-dist d))))))
         id-tree)
        (if (nil? closest-id)
            (throw 'no-candidate)
            closest-id))))
