;; -*- geiser-scheme-implementation: guile -*-

;; gfx.scm - This file is part of Guile TD
;; Copyright (C) 2017  Christopher Howard

;; Guile TD is free software: you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; Guile TD is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (td gfx)
  #:use-module (aa-tree)
  #:use-module (chickadee)
  #:use-module (chickadee math vector)
  #:use-module (chickadee render font)
  #:use-module (chickadee render sprite)
  #:use-module (chickadee render texture)
  #:use-module (srfi srfi-1)
  #:export (font-gns
           center-point
           load-field
           load-atlases
           draw-tile
           load-sprites
           draw-frame-rate
           draw-upgrade-hint
           field
           draw-tower-hint))

(read-set! keywords 'prefix)

(define atlases #nil)
(define field #nil)
(define font-gns #nil)
(define frame-time-before 30)
(define frame-time-after 30)
(define frame-rate 30)
(define frame-rate-prev 30)
(define center-point (vec2 512 384))

(define (load-field)
  (set! field
    (load-image "maps/rt-based/map0.png")))

(define (load-atlases)
  (set! atlases
    (list-to-aa-tree atlases
                     (list #:t40
                           (split-texture (load-image "images/early/40.png")
                                          40 40)
                           #:t64
                           (split-texture (load-image "images/cmh/towers64.png")
                                          64 64)
                           #:s40
                           (split-texture (load-image "images/cmh/soldier40.png")
                                          40 40)
                           #:v64
                           (split-texture (load-image "images/cmh/vehicles64.png")
                                          64 64)
                           #:v64b
                           (split-texture (load-image "images/cmh/vehicles64b.png")
                                          64 64)
                           #:defeat
                           (split-texture (load-image "images/cmh/defeat.png")
                                          400 400)
                           #:victory
                           (split-texture (load-image "images/cmh/victory.png")
                                          400 400)))))

(define (draw-tile key position)
  (let* ((r (aa-search texture-mapping-tree key))
         (atlas-key (first r))
         (tile (second r))
         (atlas (aa-search atlases atlas-key))
         (tex (texture-atlas-ref atlas tile))
         (offset (third r)))
    (draw-sprite tex position #:origin offset)))

(define texture-mapping-tree
  (list-to-aa-tree
   #nil
   `(#:slow-infantry (_ #:t40 0 ,(vec2 20 20))
     #:basic-turret (_ #:t64 0 ,(vec2 32 32))
     #:basic-turret-l2 (_ #:t64 1 ,(vec2 32 32))
     #:basic-turret-l3 (_ #:t64 2 ,(vec2 32 32))
     #:delay-turret (_ #:t64 3 ,(vec2 32 32))
     #:delay-turret-l2 (_ #:t64 4 ,(vec2 32 32))
     #:delay-turret-l3 (_ #:t64 5 ,(vec2 32 32))
     #:resource-turret (_ #:t40 3 ,(vec2 20 20))
     #:resource-turret-l2 (_ #:t64 7 ,(vec2 32 32))
     #:resource-turret-l3 (_ #:t64 8 ,(vec2 32 32))
     #:resource-turret-ghost (_ #:t64 9 ,(vec2 32 32))
     #:basic-turret-ghost (_ #:t64 10 ,(vec2 32 32))
     #:delay-turret-ghost (_ #:t64 11 ,(vec2 32 32))
     #:bullet (_ #:t40 2 ,(vec2 20 20))
     #:fire-bullet (_ #:t40 1 ,(vec2 20 20))
     #:white-bullet (_ #:t40 5 ,(vec2 20 20))
     #:purple-bullet (_ #:t40 6 ,(vec2 20 20))
     #:vehicle-west (_ #:a40 0 ,(vec2 20 20))
     #:vehicle-west-blink (_ #:a40 1 ,(vec2 20 20))
     #:vehicle-south (_ #:a40 2 ,(vec2 20 20))
     #:vehicle-south-blink (_ #:a40 3 ,(vec2 20 20))
     #:vehicle-east (_ #:a40 5 ,(vec2 20 20))
     #:vehicle-east-blink (_ #:a40 4 ,(vec2 20 20))
     #:vehicle-north (_ #:a40 7 ,(vec2 20 20))
     #:vehicle-north-blink (_ #:a40 6 ,(vec2 20 20))
     #:soldier-n0 (_ #:s40 0 ,(vec2 20 20))
     #:soldier-n1 (_ #:s40 1 ,(vec2 20 20))
     #:soldier-n2 (_ #:s40 2 ,(vec2 20 20))
     #:soldier-e0 (_ #:s40 3 ,(vec2 20 20))
     #:soldier-e1 (_ #:s40 4 ,(vec2 20 20))
     #:soldier-e2 (_ #:s40 5 ,(vec2 20 20))
     #:soldier-w0 (_ #:s40 6 ,(vec2 20 20))
     #:soldier-w1 (_ #:s40 7 ,(vec2 20 20))
     #:soldier-w2 (_ #:s40 8 ,(vec2 20 20))
     #:soldier-s0 (_ #:s40 9 ,(vec2 20 20))
     #:soldier-s1 (_ #:s40 10 ,(vec2 20 20))
     #:soldier-s2 (_ #:s40 11 ,(vec2 20 20))
     #:pt-boat-east (_ #:v64 8 ,(vec2 32 32))
     #:pt-boat-south (_ #:v64 9 ,(vec2 32 32))
     #:pt-boat-west (_ #:v64 10 ,(vec2 32 32))
     #:pt-boat-north (_ #:v64 11 ,(vec2 32 32))
     #:destroyer-east (_ #:v64 12 ,(vec2 32 32))
     #:destroyer-south (_ #:v64 13 ,(vec2 32 32))
     #:destroyer-west (_ #:v64 14 ,(vec2 32 32))
     #:destroyer-north (_ #:v64 15 ,(vec2 32 32))
     #:tank-east (_ #:v64 4 ,(vec2 32 32))
     #:tank-south (_ #:v64 5 ,(vec2 32 32))
     #:tank-west (_ #:v64 6 ,(vec2 32 32))
     #:tank-north (_ #:v64 7 ,(vec2 32 32))
     #:tracked-soldier-east (_ #:v64b 0 ,(vec2 32 32))
     #:tracked-soldier-south (_ #:v64b 1 ,(vec2 32 32))
     #:tracked-soldier-west (_ #:v64b 2 ,(vec2 32 32))
     #:tracked-soldier-north (_ #:v64b 3 ,(vec2 32 32))
     #:helo-east (_ #:v64b 4 ,(vec2 32 32))
     #:helo-south (_ #:v64b 5 ,(vec2 32 32))
     #:helo-west (_ #:v64b 6 ,(vec2 32 32))
     #:helo-north (_ #:v64b 7 ,(vec2 32 32))
     #:three-wheeler-east (_ #:v64 0 ,(vec2 32 32))
     #:three-wheeler-south (_ #:v64 1 ,(vec2 32 32))
     #:three-wheeler-west (_ #:v64 2 ,(vec2 32 32))
     #:three-wheeler-north (_ #:v64 3 ,(vec2 32 32))
     #:defeat (_ #:defeat 0 ,(vec2 200 200))
     #:victory (_ #:victory 0 ,(vec2 200 200))
     )))

(define (load-sprites)
  (load-field)
  (load-atlases))

(add-hook! load-hook (lambda () (load-sprites)))

(add-hook! load-hook
           (lambda ()
             (set! font-gns
               (load-font "font/gn/good_neighbors_starling.xml"))))
    
(add-hook! before-draw-hook
           (lambda ()
             (set! frame-time-before frame-time-after)
             (set! frame-time-after (tms:clock (times)))
             (set! frame-rate-prev frame-rate)
             (set! frame-rate
               (catch 'numerical-overflow
                 (lambda ()
                   (+ (* frame-rate-prev 0.95)
                      (* (/ 1000000000.0
                            (- frame-time-after frame-time-before))
                         0.05)))
                 (lambda _ 99)))))

(define (draw-frame-rate label-vec digits-vec)
  (draw-text font-gns "FPS#: " label-vec)
  (draw-text font-gns
             (number->string (inexact->exact (round frame-rate)))
             digits-vec))
  
(define (draw-upgrade-hint center-coords-vec)
  (draw-text font-gns "upgrade!" center-coords-vec))

(define (draw-tower-hint tower-selection center-coords-vec)
  (draw-tile
   (case tower-selection
     ((#:resource-turret) #:resource-turret-ghost)
     ((#:basic-turret) #:basic-turret-ghost)
     ((#:delay-turret) #:delay-turret-ghost))
   center-coords-vec))
